@@ -48,6 +48,7 @@ USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include <xf86drm.h>
 #include <i915_drm.h>
 #include <dri2.h>
+#include <compositeext.h>
 
 #if DRI2INFOREC_VERSION <= 2
 #error DRI2 version supported by the Xserver is too old
@@ -61,12 +62,12 @@ USE OR OTHER DEALINGS IN THE SOFTWARE.
 #define COLOR_PREFER_TILING_Y 0
 
 enum frame_event_type {
+	DRI2_WAITMSC = 0,
 	DRI2_SWAP,
 	DRI2_SWAP_WAIT,
 	DRI2_SWAP_THROTTLE,
 	DRI2_FLIP,
 	DRI2_FLIP_THROTTLE,
-	DRI2_WAITMSC,
 };
 
 struct sna_dri_frame_event {
@@ -286,6 +287,8 @@ sna_dri_create_buffer(DrawablePtr draw,
 		if (bo == NULL)
 			return NULL;
 
+		assert(sna_pixmap(pixmap) != NULL);
+
 		bo = ref(bo);
 		bpp = pixmap->drawable.bitsPerPixel;
 		if (pixmap == sna->front)
@@ -298,14 +301,15 @@ sna_dri_create_buffer(DrawablePtr draw,
 		break;
 
 	case DRI2BufferBackLeft:
+		if (draw->width  == sna->front->drawable.width &&
+		    draw->height == sna->front->drawable.height &&
+		    (sna->flags & SNA_NO_FLIP) == 0)
+			flags |= CREATE_SCANOUT;
 	case DRI2BufferBackRight:
 	case DRI2BufferFrontRight:
 	case DRI2BufferFakeFrontLeft:
 	case DRI2BufferFakeFrontRight:
 		bpp = draw->bitsPerPixel;
-		if (draw->width  == sna->front->drawable.width &&
-		    draw->height == sna->front->drawable.height)
-			flags |= CREATE_SCANOUT;
 		DBG(("%s: creating back buffer %dx%d, suitable for scanout? %d\n",
 		     __FUNCTION__,
 		     draw->width, draw->height,
@@ -482,19 +486,9 @@ static void sna_dri_reference_buffer(DRI2Buffer2Ptr buffer)
 	get_private(buffer)->refcnt++;
 }
 
-static void damage(PixmapPtr pixmap, RegionPtr region)
+static void damage(PixmapPtr pixmap, struct sna_pixmap *priv, RegionPtr region)
 {
-	struct sna_pixmap *priv;
-
-	priv = sna_pixmap(pixmap);
-	assert(priv != NULL);
 	assert(priv->gpu_bo);
-
-	if (priv->cow) {
-		sna_pixmap_undo_cow(to_sna_from_pixmap(pixmap), priv,
-				    region ? MOVE_READ : 0);
-	}
-
 	if (DAMAGE_IS_ALL(priv->gpu_damage))
 		return;
 
@@ -588,16 +582,16 @@ static void sna_dri_select_mode(struct sna *sna, struct kgem_bo *dst, struct kge
 	}
 
 	VG_CLEAR(busy);
-	busy.handle = dst->handle;
+	busy.handle = src->handle;
 	if (drmIoctl(sna->kgem.fd, DRM_IOCTL_I915_GEM_BUSY, &busy))
 		return;
 
-	DBG(("%s: dst busy?=%x\n", __FUNCTION__, busy.busy));
+	DBG(("%s: src busy?=%x\n", __FUNCTION__, busy.busy));
 	if (busy.busy == 0) {
-		busy.handle = src->handle;
+		busy.handle = dst->handle;
 		if (drmIoctl(sna->kgem.fd, DRM_IOCTL_I915_GEM_BUSY, &busy))
 			return;
-		DBG(("%s: src busy?=%x\n", __FUNCTION__, busy.busy));
+		DBG(("%s: dst busy?=%x\n", __FUNCTION__, busy.busy));
 		if (busy.busy == 0) {
 			DBG(("%s: src/dst is idle, using defaults\n", __FUNCTION__));
 			return;
@@ -622,7 +616,7 @@ static void sna_dri_select_mode(struct sna *sna, struct kgem_bo *dst, struct kge
 	mode = KGEM_RENDER;
 	if (busy.busy & (1 << 17))
 		mode = KGEM_BLT;
-	kgem_bo_mark_busy(dst, mode);
+	kgem_bo_mark_busy(busy.handle == src->handle ? src : dst, mode);
 	_kgem_set_mode(&sna->kgem, mode);
 }
 
@@ -672,20 +666,15 @@ __sna_dri_copy_region(struct sna *sna, DrawablePtr draw, RegionPtr region,
 	 * the risk of copying around invalid data. So either you may not
 	 * see the results of the copy, or you may see the wrong pixels.
 	 * Either way you eventually lose.
+	 *
+	 * We also have to be careful in case that the stale buffers are
+	 * now attached to invalid (non-DRI) pixmaps.
 	 */
 
 	assert(dst->attachment == DRI2BufferFrontLeft ||
 	       src->attachment == DRI2BufferFrontLeft);
 	assert(dst->attachment != src->attachment);
 
-	src_bo = src_priv->bo;
-	if (src->attachment == DRI2BufferFrontLeft)
-		src_bo = sna_pixmap_get_bo(pixmap);
-
-	dst_bo = dst_priv->bo;
-	if (dst->attachment == DRI2BufferFrontLeft)
-		dst_bo = sna_pixmap_get_bo(pixmap);
-
 	/* Copy the minimum of the Drawable / src / dst extents */
 	w = draw->width;
 	if ((src_priv->size & 0xffff) < w)
@@ -717,16 +706,6 @@ __sna_dri_copy_region(struct sna *sna, DrawablePtr draw, RegionPtr region,
 		return NULL;
 	}
 
-	if (!wedged(sna)) {
-		if (dst->attachment != DRI2BufferFrontLeft)
-			sync = false;
-		if (sync)
-			sync = sna_pixmap_is_scanout(sna, pixmap);
-
-		sna_dri_select_mode(sna, dst_bo, src_bo, sync);
-	} else
-		sync = false;
-
 	sx = sy = dx = dy = 0;
 	if (dst->attachment == DRI2BufferFrontLeft) {
 		sx = -draw->x;
@@ -756,15 +735,6 @@ __sna_dri_copy_region(struct sna *sna, DrawablePtr draw, RegionPtr region,
 			region = &clip;
 		}
 
-		if (sync) {
-			xf86CrtcPtr crtc;
-
-			crtc = sna_covering_crtc(sna->scrn, &clip.extents, NULL);
-			sync = crtc &&
-				sna_wait_for_scanline(sna, pixmap, crtc,
-						&clip.extents);
-		}
-
 		if (get_drawable_deltas(draw, pixmap, &tx, &ty)) {
 			if (dst->attachment == DRI2BufferFrontLeft) {
 				pixman_region_translate(region ?: &clip, tx, ty);
@@ -775,9 +745,52 @@ __sna_dri_copy_region(struct sna *sna, DrawablePtr draw, RegionPtr region,
 				sy += ty;
 			}
 		}
+	} else
+		sync = false;
+
+	src_bo = src_priv->bo;
+	if (src->attachment == DRI2BufferFrontLeft) {
+		struct sna_pixmap *priv;
+
+		priv = sna_pixmap_move_to_gpu(pixmap, MOVE_READ);
+		if (priv)
+			src_bo = priv->gpu_bo;
 	}
-	if (dst->attachment == DRI2BufferFrontLeft)
-		damage(pixmap, region);
+
+	dst_bo = dst_priv->bo;
+	if (dst->attachment == DRI2BufferFrontLeft) {
+		struct sna_pixmap *priv;
+		unsigned int flags;
+
+		flags = MOVE_WRITE | __MOVE_FORCE;
+		if (clip.data ||
+		    clip.extents.x1 > 0 ||
+		    clip.extents.x2 < pixmap->drawable.width ||
+		    clip.extents.y1 > 0 ||
+		    clip.extents.y2 < pixmap->drawable.height)
+			flags |= MOVE_READ;
+
+		priv = sna_pixmap_move_to_gpu(pixmap, flags);
+		if (priv) {
+			damage(pixmap, priv, region);
+			dst_bo = priv->gpu_bo;
+		}
+	} else
+		sync = false;
+
+	if (!wedged(sna)) {
+		xf86CrtcPtr crtc;
+
+		crtc = NULL;
+		if (sync && sna_pixmap_is_scanout(sna, pixmap))
+			crtc = sna_covering_crtc(sna, &clip.extents, NULL);
+		sna_dri_select_mode(sna, dst_bo, src_bo, crtc != NULL);
+
+		sync = (crtc != NULL&&
+			sna_wait_for_scanline(sna, pixmap, crtc,
+					      &clip.extents));
+	}
+
 	if (region) {
 		boxes = REGION_RECTS(region);
 		n = REGION_NUM_RECTS(region);
@@ -788,6 +801,7 @@ __sna_dri_copy_region(struct sna *sna, DrawablePtr draw, RegionPtr region,
 		n = 1;
 	}
 	DamageRegionAppend(&pixmap->drawable, region);
+
 	if (wedged(sna)) {
 		sna_dri_copy_fallback(sna, draw->bitsPerPixel,
 				      src_bo, sx, sy,
@@ -873,22 +887,22 @@ static inline int sna_wait_vblank(struct sna *sna, drmVBlank *vbl)
 #if DRI2INFOREC_VERSION >= 4
 
 static int
-sna_dri_get_pipe(DrawablePtr pDraw)
+sna_dri_get_pipe(DrawablePtr draw)
 {
-	ScrnInfoPtr pScrn = xf86ScreenToScrn(pDraw->pScreen);
+	struct sna *sna = to_sna_from_drawable(draw);
 	xf86CrtcPtr crtc;
 	BoxRec box;
 	int pipe;
 
-	if (pDraw->type == DRAWABLE_PIXMAP)
+	if (draw->type == DRAWABLE_PIXMAP)
 		return -1;
 
-	box.x1 = pDraw->x;
-	box.y1 = pDraw->y;
-	box.x2 = box.x1 + pDraw->width;
-	box.y2 = box.y1 + pDraw->height;
+	box.x1 = draw->x;
+	box.y1 = draw->y;
+	box.x2 = box.x1 + draw->width;
+	box.y2 = box.y1 + draw->height;
 
-	crtc = sna_covering_crtc(pScrn, &box, NULL);
+	crtc = sna_covering_crtc(sna, &box, NULL);
 
 	/* Make sure the CRTC is valid and this is the real front buffer */
 	pipe = -1;
@@ -1032,6 +1046,11 @@ sna_dri_page_flip(struct sna *sna, struct sna_dri_frame_event *info)
 	if (!info->count)
 		return false;
 
+	DBG(("%s: mark handle=%d as scanout, swap front (handle=%d, name=%d) and back (handle=%d, name=%d)\n",
+	     __FUNCTION__, bo->handle,
+	     get_private(info->front)->bo->handle, info->front->name,
+	     get_private(info->back)->bo->handle, info->back->name));
+
 	info->scanout[1] = info->scanout[0];
 	info->scanout[0].bo = ref(bo);
 	info->scanout[0].name = info->back->name;
@@ -1108,11 +1127,6 @@ can_flip(struct sna * sna,
 		return false;
 	}
 
-	if (!get_private(front)->scanout) {
-		DBG(("%s: no, DRI2 drawable not attached at time of creation)\n",
-		     __FUNCTION__));
-		return false;
-	}
 	assert(get_private(front)->pixmap == sna->front);
 	assert(sna_pixmap(sna->front)->gpu_bo == get_private(front)->bo);
 
@@ -1408,7 +1422,7 @@ sna_dri_immediate_blit(struct sna *sna,
 			ret = true;
 	} else {
 		info->bo = __sna_dri_copy_region(sna, draw, NULL,
-						 info->back, info->front, true);
+						 info->back, info->front, false);
 		if (event)
 			DRI2SwapComplete(info->client, draw, 0, 0, 0,
 					 DRI2_BLIT_COMPLETE,
@@ -1582,8 +1596,8 @@ static void sna_dri_flip_event(struct sna *sna,
 	/* We assume our flips arrive in order, so we don't check the frame */
 	switch (flip->type) {
 	case DRI2_FLIP:
-		DBG(("%s: flip complete (drawable gone? %d)\n",
-		     __FUNCTION__, flip->draw == NULL));
+		DBG(("%s: flip complete (drawable gone? %d), msc=%d\n",
+		     __FUNCTION__, flip->draw == NULL, flip->fe_frame));
 		if (flip->draw)
 			DRI2SwapComplete(flip->client, flip->draw,
 					 flip->fe_frame,
@@ -1675,6 +1689,34 @@ get_current_msc_for_target(struct sna *sna, CARD64 target_msc, int pipe)
 	return ret;
 }
 
+static Bool find(pointer value, XID id, pointer cdata)
+{
+	return TRUE;
+}
+
+static int use_triple_buffer(struct sna *sna, ClientPtr client)
+{
+	struct sna_client *priv;
+
+	if ((sna->flags & SNA_TRIPLE_BUFFER) == 0)
+		return DRI2_FLIP;
+
+	/* Hack: Disable triple buffering for compositors */
+
+#if XORG_VERSION_CURRENT >= XORG_VERSION_NUMERIC(1,12,99,901,0)
+	priv = sna_client(client);
+	if (priv->is_compositor == 0)
+		priv->is_compositor =
+			LookupClientResourceComplex(client,
+						    CompositeClientWindowType+1,
+						    find, NULL) ? DRI2_FLIP : DRI2_FLIP_THROTTLE;
+
+	return priv->is_compositor;
+#else
+	return DRI2_FLIP_THROTTLE;
+#endif
+}
+
 static bool
 sna_dri_schedule_flip(ClientPtr client, DrawablePtr draw,
 		      DRI2BufferPtr front, DRI2BufferPtr back, int pipe,
@@ -1697,8 +1739,7 @@ sna_dri_schedule_flip(ClientPtr client, DrawablePtr draw,
 		DBG(("%s: performing immediate swap on pipe %d, pending? %d, mode: %d\n",
 		     __FUNCTION__, pipe, info != NULL, info ? info->mode : 0));
 
-		if (info &&
-		    info->draw == draw) {
+		if (info && info->draw == draw) {
 			assert(info->type == DRI2_FLIP_THROTTLE);
 			assert(info->front == front);
 			if (info->back != back) {
@@ -1724,8 +1765,7 @@ sna_dri_schedule_flip(ClientPtr client, DrawablePtr draw,
 		if (info == NULL)
 			return false;
 
-		info->type = sna->flags & SNA_TRIPLE_BUFFER ? DRI2_FLIP_THROTTLE: DRI2_FLIP;
-
+		info->type = use_triple_buffer(sna, client);
 		info->draw = draw;
 		info->client = client;
 		info->event_complete = func;
@@ -1751,17 +1791,15 @@ sna_dri_schedule_flip(ClientPtr client, DrawablePtr draw,
 			     __FUNCTION__));
 			info->type = DRI2_FLIP;
 			sna->dri.flip_pending = info;
-			*target_msc = current_msc + 1;
-			return true;
-		}
-
-		if (!sna_dri_page_flip(sna, info)) {
-			sna_dri_frame_event_info_free(sna, draw, info);
-			return false;
+		} else {
+			if (!sna_dri_page_flip(sna, info)) {
+				sna_dri_frame_event_info_free(sna, draw, info);
+				return false;
+			}
 		}
 
+		current_msc++;
 		if (info->type != DRI2_FLIP) {
-			current_msc++;
 new_back:
 			sna_dri_flip_get_back(sna, info);
 			DRI2SwapComplete(client, draw, 0, 0, 0,
@@ -1769,6 +1807,7 @@ new_back:
 					 func, data);
 		}
 out:
+		DBG(("%s: target_msc=%lu\n", __FUNCTION__, (unsigned long)current_msc));
 		*target_msc = current_msc;
 		return true;
 	}
@@ -2048,33 +2087,6 @@ skip:
 	return TRUE;
 }
 
-#if USE_ASYNC_SWAP
-static Bool
-sna_dri_async_swap(ClientPtr client, DrawablePtr draw,
-		   DRI2BufferPtr front, DRI2BufferPtr back,
-		   DRI2SwapEventPtr func, void *data)
-{
-	struct sna *sna = to_sna_from_drawable(draw);
-	CARD64 target_msc = 0;
-	int pipe;
-
-	DBG(("%s()\n", __FUNCTION__));
-
-	if (!can_flip(sna, draw, front, back) ||
-	    (pipe = sna_dri_get_pipe(draw)) < 0 ||
-	    !sna_dri_schedule_flip(client, draw, front, back, pipe,
-				   &target_msc, 0, 0, func, data)) {
-		DBG(("%s: unable to flip, so blit\n", __FUNCTION__));
-		__sna_dri_copy_region(sna, draw, NULL, back, front, false);
-
-		DRI2SwapComplete(client, draw, 0, 0, 0,
-				 DRI2_BLIT_COMPLETE, func, data);
-		return false;
-	}
-	return TRUE;
-}
-#endif
-
 static uint64_t gettime_us(void)
 {
 	struct timespec tv;
@@ -2288,7 +2300,7 @@ bool sna_dri_open(struct sna *sna, ScreenPtr screen)
 	memset(&info, '\0', sizeof(info));
 	info.fd = sna->kgem.fd;
 	info.driverName = dri_driver_name(sna);
-	info.deviceName = intel_get_device_name(sna->scrn);
+	info.deviceName = intel_get_client_name(sna->scrn);
 
 	DBG(("%s: loading dri driver '%s' [gen=%d] for device '%s'\n",
 	     __FUNCTION__, info.driverName, sna->kgem.gen, info.deviceName));
@@ -2316,7 +2328,7 @@ bool sna_dri_open(struct sna *sna, ScreenPtr screen)
 
 #if USE_ASYNC_SWAP
 	info.version = 10;
-	info.AsyncSwap = sna_dri_async_swap;
+	info.scheduleSwap0 = 1;
 #endif
 
 	return DRI2ScreenInit(screen, &info);
